@@ -4,7 +4,8 @@ from collections import defaultdict
 from datetime import datetime, timezone
 from decimal import Decimal
 from enum import Enum
-from typing import Any, AsyncGenerator, DefaultDict, Dict, List, NoReturn, Optional, Set, Tuple, cast
+from functools import wraps
+from typing import Any, AsyncGenerator, Callable, DefaultDict, Dict, List, NoReturn, Optional, Set, Tuple, cast
 
 from aiohttp import ClientResponseError
 from aiosignalrcore.hub.base_hub_connection import BaseHubConnection  # type: ignore
@@ -16,13 +17,13 @@ from dipdup.config import (
     BigMapIndexConfig,
     ContractConfig,
     HTTPConfig,
-    IndexConfigTemplateT,
     OperationHandlerOriginationPatternConfig,
     OperationIndexConfig,
+    ResolvedIndexConfigT,
 )
 from dipdup.datasources.datasource import IndexDatasource
 from dipdup.datasources.tzkt.enums import TzktMessageType
-from dipdup.models import BigMapAction, BigMapData, BlockData, HeadBlockData, OperationData
+from dipdup.models import BigMapAction, BigMapData, BlockData, Head, HeadBlockData, OperationData, QuoteData
 from dipdup.utils import groupby, split_by_chunks
 
 OperationID = int
@@ -279,6 +280,38 @@ class BigMapFetcher:
             yield big_maps[0].level, big_maps[: i + 1]
 
 
+class BlockCache:
+    def __init__(self) -> None:
+        # FIXME: Why store older blocks?
+        self._limit = 10
+        self._blocks: DefaultDict[int, Optional[HeadBlockData]] = defaultdict(lambda: None)
+        self._events: DefaultDict[int, asyncio.Event] = defaultdict(asyncio.Event)
+
+    async def add_block(self, block: HeadBlockData) -> None:
+        if self._blocks[block.level]:
+            raise RuntimeError('Attempt to add block {block.level} which is already cached')
+        self._blocks[block.level] = block
+        self._events[block.level].set()
+
+        last_blocks = sorted(self._blocks.keys())[-self._limit :]
+        self._blocks = defaultdict(lambda: None, ({k: v for k, v in self._blocks.items() if k in last_blocks}))
+        self._events = defaultdict(asyncio.Event, ({k: v for k, v in self._events.items() if k in last_blocks}))
+
+    async def get_block(self, level: int) -> HeadBlockData:
+        if self._blocks and level < sorted(self._blocks.keys())[-1] - self._limit:
+            raise RuntimeError(f'Attemps to get block older than {self._limit} levels from head')
+
+        try:
+            await asyncio.wait_for(fut=self._events[level].wait(), timeout=3)
+        except asyncio.TimeoutError as e:
+            raise RuntimeError(f'Block {level} hasn\'t arrived in 3 seconds. Forgot to subscribe to head?') from e
+
+        block = self._blocks[level]
+        if not block:
+            raise RuntimeError('Event is set but block is missing in cache')
+        return block
+
+
 class TzktDatasource(IndexDatasource):
     """Bridge between REST/WS TzKT endpoints and DipDup.
 
@@ -308,9 +341,10 @@ class TzktDatasource(IndexDatasource):
         self._transaction_subscriptions: Set[str] = set()
         self._origination_subscriptions: bool = False
         self._big_map_subscriptions: Dict[str, Set[str]] = {}
-        self._client: Optional[BaseHubConnection] = None
+        self._ws_client: Optional[BaseHubConnection] = None
 
-        self._block: Optional[HeadBlockData] = None
+        self._block_cache: BlockCache = BlockCache()
+        self._head: Optional[Head] = None
         self._level: Optional[int] = None
         self._sync_level: Optional[int] = None
 
@@ -327,10 +361,8 @@ class TzktDatasource(IndexDatasource):
         return self._sync_level
 
     @property
-    def block(self) -> HeadBlockData:
-        if self._block is None:
-            raise RuntimeError('Attempt to access head block before the first message')
-        return self._block
+    def head(self) -> Optional[Head]:
+        return self._head
 
     async def get_similar_contracts(self, address: str, strict: bool = False) -> List[str]:
         """Get list of contracts sharing the same code hash or type hash"""
@@ -506,7 +538,33 @@ class TzktDatasource(IndexDatasource):
             big_maps.append(self.convert_big_map(bm))
         return big_maps
 
-    async def add_index(self, index_config: IndexConfigTemplateT) -> None:
+    async def get_quote(self, level: int) -> QuoteData:
+        """Get quote for block"""
+        self._logger.info('Fetching quotes for level %s', level)
+        quote_json = await self._http.request(
+            'get',
+            url='v1/quotes',
+            params={"level": level},
+            cache=True,
+        )
+        return self.convert_quote(quote_json[0])
+
+    async def get_quotes(self, from_level: int, to_level: int) -> List[QuoteData]:
+        """Get quotes for blocks"""
+        self._logger.info('Fetching quotes for levels %s-%s', from_level, to_level)
+        quotes_json = await self._http.request(
+            'get',
+            url='v1/quotes',
+            params={
+                "level.ge": from_level,
+                "level.lt": to_level,
+                "limit": self.request_limit,
+            },
+            cache=False,
+        )
+        return [self.convert_quote(quote) for quote in quotes_json]
+
+    async def add_index(self, index_config: ResolvedIndexConfigT) -> None:
         """Register index config in internal mappings and matchers. Find and register subscriptions."""
 
         if isinstance(index_config, OperationIndexConfig):
@@ -530,45 +588,56 @@ class TzktDatasource(IndexDatasource):
 
         await self._on_connect()
 
-    def _get_client(self) -> BaseHubConnection:
+    def _get_ws_client(self) -> BaseHubConnection:
         """Create SignalR client, register message callbacks"""
-        if self._client is None:
-            self._logger.info('Creating websocket client')
-            self._client = (
-                HubConnectionBuilder()
-                .with_url(self._http._url + '/v1/events')
-                .with_automatic_reconnect(
-                    {
-                        "type": "raw",
-                        "keep_alive_interval": 10,
-                        "reconnect_interval": 5,
-                        "max_attempts": 5,
-                    }
-                )
-            ).build()
+        if self._ws_client:
+            return self._ws_client
 
-            self._client.on_open(self._on_connect)
-            self._client.on_error(self._on_error)
-            self._client.on('operations', self._on_operation_message)
-            self._client.on('bigmaps', self._on_big_map_message)
-            self._client.on('head', self._on_head_message)
+        self._logger.info('Creating websocket client')
+        self._ws_client = (
+            HubConnectionBuilder()
+            .with_url(self._http._url + '/v1/events')
+            .with_automatic_reconnect(
+                {
+                    "type": "raw",
+                    "keep_alive_interval": 10,
+                    "reconnect_interval": 5,
+                    "max_attempts": 5,
+                }
+            )
+        ).build()
 
-        return self._client
+        _ws_lock = asyncio.Lock()
+
+        def _lock_wrapper(fn: Callable):
+            @wraps(fn)
+            async def _wrapper(*args, **kwargs):
+                async with _ws_lock:
+                    return await fn(*args, **kwargs)
+
+            return _wrapper
+
+        self._ws_client.on_open(_lock_wrapper(self._on_connect))
+        self._ws_client.on_error(_lock_wrapper(self._on_error))
+        self._ws_client.on('operations', _lock_wrapper(self._on_operation_message))
+        self._ws_client.on('bigmaps', _lock_wrapper(self._on_big_map_message))
+        self._ws_client.on('head', _lock_wrapper(self._on_head_message))
+
+        return self._ws_client
 
     async def run(self) -> None:
         """Main loop. Sync indexes via REST, start WS connection"""
-        # TODO: Update docs, honor `realtime` flag
         self._logger.info('Starting datasource')
 
         self._logger.info('Starting websocket client')
-        await self._get_client().start()
+        await self._get_ws_client().start()
 
     async def _on_connect(self) -> None:
         """Subscribe to all required channels on established WS connection"""
-        if self._get_client().transport.state != ConnectionState.connected:
+        if self._get_ws_client().transport.state != ConnectionState.connected:
             return
 
-        self._logger.info('Connected to server')
+        self._logger.info('Realtime connection established, subscribing to channels')
         await self._subscribe_to_head()
         for address in self._transaction_subscriptions:
             await self._subscribe_to_transactions(address)
@@ -587,7 +656,7 @@ class TzktDatasource(IndexDatasource):
 
     async def _subscribe_to_transactions(self, address: str) -> None:
         """Subscribe to contract's operations on established WS connection"""
-        self._logger.info('Subscribing to %s transactions', address)
+        self._logger.debug('Subscribing to %s transactions', address)
         await self._send(
             'SubscribeToOperations',
             [
@@ -600,7 +669,7 @@ class TzktDatasource(IndexDatasource):
 
     async def _subscribe_to_originations(self) -> None:
         """Subscribe to all originations on established WS connection"""
-        self._logger.info('Subscribing to originations')
+        self._logger.debug('Subscribing to originations')
         await self._send(
             'SubscribeToOperations',
             [
@@ -612,7 +681,7 @@ class TzktDatasource(IndexDatasource):
 
     async def _subscribe_to_big_maps(self, address: str, paths: Set[str]) -> None:
         """Subscribe to contract's big map diffs on established WS connection"""
-        self._logger.info('Subscribing to big map updates of %s, %s', address, paths)
+        self._logger.debug('Subscribing to big map updates of %s, %s', address, paths)
         for path in paths:
             await self._send(
                 'SubscribeToBigMaps',
@@ -626,18 +695,22 @@ class TzktDatasource(IndexDatasource):
 
     async def _subscribe_to_head(self) -> None:
         """Subscribe to head on established WS connection"""
-        self._logger.info('Subscribing to head')
+        self._logger.debug('Subscribing to head')
         await self._send(
             'SubscribeToHead',
             [],
         )
 
-    async def _extract_message_data(self, channel: str, message: List[Any]) -> Any:
+    async def _extract_message_data(self, channel: str, message: List[Any]) -> AsyncGenerator[Tuple[int, Dict], None]:
         for item in message:
-            head_level = item['state']
             message_type = TzktMessageType(item['type'])
             self._logger.debug('`%s` message: %s', channel, message_type.name)
 
+            head_level = item['state']
+            if self._level and head_level < self._level:
+                raise RuntimeError(f'Received data message from level lower than current: {head_level} < {self._level}')
+
+            # NOTE: State messages will be replaced with WS negotiation some day
             if message_type == TzktMessageType.STATE:
                 if self._sync_level != head_level:
                     self._logger.info('Datasource level set to %s', head_level)
@@ -646,11 +719,11 @@ class TzktDatasource(IndexDatasource):
 
             elif message_type == TzktMessageType.DATA:
                 self._level = head_level
-                yield item['data']
+                yield head_level, item['data']
 
             elif message_type == TzktMessageType.REORG:
                 if self.level is None:
-                    raise RuntimeError
+                    raise RuntimeError('Reorg message received but datasource is not connected')
                 self.emit_rollback(self.level, head_level)
 
             else:
@@ -658,7 +731,8 @@ class TzktDatasource(IndexDatasource):
 
     async def _on_operation_message(self, message: List[Dict[str, Any]]) -> None:
         """Parse and emit raw operations from WS"""
-        async for data in self._extract_message_data('operation', message):
+        async for level, data in self._extract_message_data('operation', message):
+            block = await self._block_cache.get_block(level)
             operations = []
             for operation_json in data:
                 operation = self.convert_operation(operation_json)
@@ -666,23 +740,66 @@ class TzktDatasource(IndexDatasource):
                     continue
                 operations.append(operation)
             if operations:
-                self.emit_operations(operations, self.block)
+                self.emit_operations(operations, block)
 
     async def _on_big_map_message(self, message: List[Dict[str, Any]]) -> None:
         """Parse and emit raw big map diffs from WS"""
-        async for data in self._extract_message_data('big_map', message):
+        async for level, data in self._extract_message_data('big_map', message):
+            block = await self._block_cache.get_block(level)
             big_maps = []
             for big_map_json in data:
                 big_map = self.convert_big_map(big_map_json)
                 big_maps.append(big_map)
-            self.emit_big_maps(big_maps, self.block)
+            self.emit_big_maps(big_maps, block)
 
     async def _on_head_message(self, message: List[Dict[str, Any]]) -> None:
-        async for data in self._extract_message_data('head', message):
+        async for _, data in self._extract_message_data('head', message):
             block = self.convert_head_block(data)
-            self._block = block
+            await self._block_cache.add_block(block)
+
+            created = False
+            if self._head is None:
+                self._head, created = await Head.get_or_create(
+                    name=self._http._url,
+                    defaults=dict(
+                        level=block.level,
+                        hash=block.hash,
+                        timestamp=block.timestamp,
+                    ),
+                )
+            if not created:
+                self._head.level = block.level  # type: ignore
+                self._head.hash = block.hash  # type: ignore
+                self._head.timestamp = block.timestamp  # type: ignore
+                await self._head.save()
+
             self.emit_head(block)
 
+    # FIXME: I don't like this approach, too hacky.
+    async def set_head_from_http(self) -> None:
+        """Set block from `get_head_block` HTTP method for indexes to use the same level during initial sync"""
+        if self._head:
+            self._logger.warning('Attempt to set head twice')
+            return
+        block = await self.get_head_block()
+        self._head, created = await Head.get_or_create(
+            name=self._http._url,
+            defaults=dict(
+                level=block.level,
+                hash=block.hash,
+                timestamp=block.timestamp,
+            ),
+        )
+        if not created:
+            self._head.level = block.level  # type: ignore
+            self._head.hash = block.hash  # type: ignore
+            self._head.timestamp = block.timestamp  # type: ignore
+            await self._head.save()
+
+        self._logger.info('Datasource head set to block with level %s', self._head.level)
+
+        # NOTE: No need to emit?
+
     @classmethod
     def convert_operation(cls, operation_json: Dict[str, Any]) -> OperationData:
         """Convert raw operation message from WS/REST into dataclass"""
@@ -759,7 +876,7 @@ class TzktDatasource(IndexDatasource):
         return BigMapData(
             id=big_map_json['id'],
             level=big_map_json['level'],
-            # FIXME: operation_id field in API
+            # FIXME: missing `operation_id` field in API to identify operation
             operation_id=big_map_json['level'],
             timestamp=cls._parse_timestamp(big_map_json['timestamp']),
             bigmap=big_map_json['bigmap'],
@@ -812,8 +929,23 @@ class TzktDatasource(IndexDatasource):
             quote_eth=Decimal(head_block_json['quoteEth']),
         )
 
+    @classmethod
+    def convert_quote(cls, quote_json: Dict[str, Any]) -> QuoteData:
+        """Convert raw quote message from REST into dataclass"""
+        return QuoteData(
+            level=quote_json['level'],
+            timestamp=cls._parse_timestamp(quote_json['timestamp']),
+            btc=Decimal(quote_json['btc']),
+            eur=Decimal(quote_json['eur']),
+            usd=Decimal(quote_json['usd']),
+            cny=Decimal(quote_json['cny']),
+            jpy=Decimal(quote_json['jpy']),
+            krw=Decimal(quote_json['krw']),
+            eth=Decimal(quote_json['eth']),
+        )
+
     async def _send(self, method: str, arguments: List[Dict[str, Any]], on_invocation=None) -> None:
-        client = self._get_client()
+        client = self._get_ws_client()
         while client.transport.state != ConnectionState.connected:
             await asyncio.sleep(0.1)
         await client.send(method, arguments, on_invocation)
